'use strict';

const JsonPatch = require('fast-json-patch');
const cassandra = require('cassandra-driver');
const Uuid = cassandra.types.Uuid;

/**
 * Validates the specified patch operations using the fast-json-patch default validation.
 * @param {*} patchOps The JSON patch operations matching RFC6902
 * @param {*} document The target JSON document.
 * @see https://tools.ietf.org/html/rfc6902 RFC6902 specification
 */
let isValidJsonPatch = function(patchOps, document) {
    var result = {};
    result = JsonPatch.validate(patchOps, document);
    if (result) {
        throw new Error(result.message);
    }
};

/**
 * Validates the specified patch operations' values have valid values.
 * @param {Object[]} patchOps The JSON patch operations matching RFC6902
 * @see https://tools.ietf.org/html/rfc6902 RFC6902 specification
 */
let hasValidValues = function(patchOps) {
    for (const patchOp of patchOps) {
        if (typeof(patchOp.value) === 'string') {
            /*
            Regex /^\s*$/:
                ^ asserts position at start of a line
                \s matches any whitespace character (equal to [\r\n\t\f\v ])
                * Quantifier — Matches between zero and unlimited times, as many times as possible, giving back as needed (greedy)
                $ asserts position at the end of a line
            */
            if (patchOp.value.match(/^\s*$/)) {
                throw new Error('Patch value must not be an empty string or whitespace.');
            }
        }
    }
};

/**
 * Validates the specified patch operations are not modifying the schema's partition keys.
 * @param {Object[]} patchOps The JSON patch operations matching RFC6902
 * @param {Object} schema The Cassandra model schema
 * @see https://tools.ietf.org/html/rfc6902 RFC6902 specification
 */
let isNotPatchingCriticalFields = function(patchOps, schema) {
    // Check for attempt to modify critical fields
    let modelKeys = schema.key;
    // the model partition key(s) are always [0]
    let modelPartitionKeys = modelKeys[0];
    // the partition key(s) could be a single string or an array of strings
    if (typeof(modelPartitionKeys) === 'string') modelPartitionKeys = [modelPartitionKeys]; 
    for(const patchOp of patchOps) {
        // the target about to be modified is the last step in the path
        let target = patchOp.path.split('/').pop();
        if (modelPartitionKeys.includes(target)) {
            throw new Error(`Invalid patch target: ${target}`);
        }
    }
};

/**
 * Validates the specified patch operations to ensure they conform to TFC6902.
 * @param {Object[]} patchOps The JSON patch operations matching RFC6902
 * @param {Object} schema The Cassandra model schema
 * @see https://tools.ietf.org/html/rfc6902 RFC6902 specification
 */
let validatePatchOps = function(patchOps, schema) {
    // Check for non-RFC compatible operations
    isValidJsonPatch(patchOps);

    hasValidValues(patchOps);

    isNotPatchingCriticalFields(patchOps, schema);
};

/**
 * Replaces the field at the given patch operation path with null.
 * @param {Object} document The target JSON document.
 * @param {String} operationPath The patch operation path.
 * @returns {*} The previous value of the field at the patch operation path.
 */
let patchRemove = function(document, operationPath) {
    let value = null;
    let step = null;
    let path = operationPath.split('/').reverse();
    path.pop(); // skip empty path element

    let pointer = document;
    while (path.length > 1) {
        step = path.pop();
        pointer = pointer[step];
    }
    step = path.pop();
    value = pointer[step];
    pointer[step] = null;

    return value;
};

/**
 * Transforms the data types of the specified document to match the given field schema.
 * @param {Object} document The target JSON document.
 * @param {Object[]} fields The field schema to apply.
 * @returns {Object} A new `Object` containing the results of the transform.
 */
let transformFields = function(document, fields) {
    let clone = Object.assign({}, document);

    for (const key in fields) {
        if (fields.hasOwnProperty(key)) {
            const field = fields[key];
            
            // convert uuid fields
            if (field.type === 'uuid') {
                clone[key] = Uuid.fromString(clone[key]);
            }
        }
    }

    return clone;
};

/**
 * Applies the specified patch operations to the given document.
 * @param {Object[]} patchOps The JSON patch operations matching RFC6902
 * @param {Object} document The target JSON document.
 * @returns {Object} A new `Object` containing the results of the patch operations. 
 * @see https://tools.ietf.org/html/rfc6902 RFC6902 specification
 */
let applyPatchOperations = function(patchOps, document) {
    /*
    The original model contains non-configurable properties which protects it from having 
    fields deleted from it.  This prevents fast-json-patch from being able to remove fields.
    To get around this the original model is converted into a string and then parsed into a 
    new JSON Object.  The fast-json-patch operations can be applied to the new JSON Object 
    which can then be used to instantiate a new model containing the updated fields.
    */
    let clone = Object.assign({}, document);

    for (const operation of patchOps) {
        /*
        During a 'remove' operation fast-json-patch attempts to delete the field at the 
        given path using `delete document[path]`.  While this would be correct for 
        modifying a JSON object, this doesn't work with our Cassandra model because its 
        properties are non-configurable.  The delete option ends up returning false without 
        throwing an exception and acts as a 'no-op'.  In order to correctly handle a 
        'remove' operation, the target field must be set to null manually.
        
        TODO: Implement Unset to avoid creating a tombstone.
        Cassandra prefers an Unset value because nullifying a field creates a tombstone.  
        Unfortunately using cassandra.types.unset instead of null causes Cassandra to 
        return errors.  I'm not sure how to address this currently.
        See:  https://docs.datastax.com/en/developer/nodejs-driver/3.2/features/datatypes/nulls/
        */
        if (operation.op === 'remove') {
            patchRemove(clone, operation.path);
        }
        /*
        The 'move' operation acts as a 'remove' followed by a 'replace'.  See the above 
        comment for 'remove' for more information.
        */
        else if (operation.op === 'move') {
            let value = patchRemove(clone, operation.from);
            clone = JsonPatch.applyOperation(clone, {op: 'replace', path: operation.path, value: value}).newDocument;
        }
        else {
            clone = JsonPatch.applyOperation(clone, operation).newDocument;
        }
    }

    return clone;
};

/**
 * Applies one or more patch operations to a specified Cassandra entity.
 * @param {Client} client The Express Cassandra client
 * @param {Object} record The Cassandra entity
 * @param {Object[]} patchOps The JSON patch operations matching RFC6902
 * @param {Object} create_instance The Cassandra model instance to create
 * @returns {Object} The patched Cassandra entity
 * @see https://tools.ietf.org/html/rfc6902 RFC6902 specification
 */
let patch = async function(client, record, patchOps, create_instance) {
    let originalRecord = Object.assign({}, record);

    validatePatchOps(patchOps, create_instance._properties.schema);

    let patchedDocument = applyPatchOperations(patchOps, record);

    // Id fields must be converted back from String to Uuid
    let fields = create_instance._properties.schema.fields;
    patchedDocument = transformFields(patchedDocument, fields);

    let patchedRecord = new create_instance(patchedDocument);

    await patchedRecord.cascadeUpdateAsync(client, originalRecord);

    return patchedRecord;
};

module.exports = {
    patch: patch,
    patchRemove: patchRemove,
    isValidJsonPatch: isValidJsonPatch,
    hasValidValues: hasValidValues,
    isNotPatchingCriticalFields: isNotPatchingCriticalFields,
    applyPatchOperations: applyPatchOperations,
    transformFields: transformFields
};