'use strict';

/**
 * Validates the specified document contains all given fields.
 * @param {Object} document The target JSON document.
 * @param {Object[]} fields The field schema.
 */
let hasRequiredFields = function(document, fields) {
    let missingFields = [];
    for (const fieldName in fields) {
        // Don't let falsey values indicate a missing field, use == null to check for
        // a truely null or undefined result instead
        if (document[fieldName] == null) {
            missingFields.push(fieldName);
        }
    }

    if (missingFields.length > 0) {
        let missingFieldsDesc = missingFields.join(', ');
        throw new Error(`All fields must be specified.  (missing fields: ${missingFieldsDesc})`);
    }
};

/**
 * Removes critical fields from the specified document for the given Cassandra model key.
 * @param {Object} document The target JSON document.
 * @param {Array[]} keys The Cassandra model key.
 * @returns {Object} A new `Object` containing no critical fields.
 */
let removeCriticalFields = function(document, key) {
    let clone = Object.assign({}, document);

    // the model partition key(s) are always [0]
    let partitionKeys = key[0];
    if (typeof(partitionKeys) === 'string') partitionKeys = [partitionKeys];
    for (const key of partitionKeys) {
        delete clone[key];
    }

    return clone;
};

/**
 * Updates the specified Cassandra entity.
 * @param {Client} client The Express Cassandra client
 * @param {Object} record The Cassandra entity
 * @param {Object} create_model The Cassandra model
 * @param {Object} document The target JSON document
 */
let update = async function(client, record, create_model) {
    let originalRecord = Object.assign({}, record);

    let fields = create_model._properties.schema.fields;

    hasRequiredFields(requestFields, fields);

    let requestFields = removeCriticalFields(requestFields, create_model._properties.schema.key);

    Object.assign(record, requestFields);

    await record.cascadeUpdateAsync(client, originalRecord);

    return record;
};

module.exports = {
    update: update,
    hasRequiredFields: hasRequiredFields,
    removeCriticalFields: removeCriticalFields
};