'use strict';

const expect = require('chai').expect;
const patch = require('../../methods/patch');
const cassandra = require('cassandra-driver');
const Uuid = cassandra.types.Uuid;

describe('patch', function () {

    let document;
    let schema;

    beforeEach(function() {
        document = {
            person_id: Uuid.random(),
            first_name: 'John',
            last_name: 'Doe',
            favorite_colors: ['blue', 'green']
        };

        schema = {
            fields: {
                person_id: { type: 'uuid', rule: { required: true } },
                first_name: { type: 'string', rule: { required: true } },
                last_name: { type: 'string', rule: { required: true } },
                favorite_colors: { type: 'array', rule: { required: false } }
            },
            key: [['last_name'], 'first_name', 'person_id']
        };
    });

    describe('patchRemove', function() {
        
        it('should replace the value at the specified path with null', function() {
            const person_id = document.person_id;
            const first_name = document.first_name;
            const last_name = document.last_name;
            const favorite_colors = document.favorite_colors;
            
            let path = '/last_name';
            
            let result = patch.patchRemove(document, path);

            expect(result).to.exist;
            expect(result).to.equal(last_name);
            
            expect(document).to.exist;
            expect(document.person_id).to.equal(person_id);
            expect(document.first_name).to.equal(first_name);
            expect(document.last_name).to.be.null;
            expect(document.favorite_colors).to.equal(favorite_colors);
        });
    });

    describe('transformFields', function() {

        it('should transform document values to uuids for the given uuid schema fields', function() {
            document.person_id = document.person_id.toString();

            let patchedDocument = patch.transformFields(document, schema.fields);

            expect(patchedDocument.person_id).to.exist;
            expect(patchedDocument.person_id).to.be.instanceOf(Uuid);
        });
    });

    describe('validation', function () {

        describe('isValidJsonPatch', function () {
            it('should validate patch sequence is an array', function () {
                let badFn = function () {
                    let patchOps = {
                        op: 'replace',
                        path: '/first_name',
                        value: 'Jane'
                    };

                    patch.isValidJsonPatch(patchOps, document);
                };

                expect(badFn).to.throw(Error, 'Patch sequence must be an array');
            });

            it('should validate operation is not an object', function () {
                let badFn = function () {
                    let patchOps = [7];

                    patch.isValidJsonPatch(patchOps, document);
                };

                expect(badFn).to.throw(Error, 'Operation is not an object');
            });

            it('should validate operation \'op\' property is not one of operations defined in RFC-6902', function () {
                let badFn = function () {
                    let patchOps = [
                        {
                            op: 'refund',
                            path: '/last_name',
                            value: 'some value'
                        }
                    ];

                    patch.isValidJsonPatch(patchOps, document);
                };

                expect(badFn).to.throw(Error, 'Operation `op` property is not one of operations defined in RFC-6902');
            });

            it('should validate operation \'path\' property is not a valid string', function () {
                let badFn = function () {
                    let patchOps = [
                        {
                            op: 'remove',
                            path: 1,
                            value: 'some value'
                        }
                    ];

                    patch.isValidJsonPatch(patchOps, document);
                };

                expect(badFn).to.throw(Error, 'Operation `path` property is not a string');
            });

            it('should validate operation \'from\' property is not present (applicable in move and copy operations)', function () {
                let badFn = function () {
                    let patchOps = [
                        {
                            op: 'move',
                            //from: '/last_name'
                            path: '/first_name'
                        }
                    ];

                    patch.isValidJsonPatch(patchOps, document);
                };

                expect(badFn).to.throw(Error, 'Operation `from` property is not present (applicable in `move` and `copy` operations');
            });

            it('should validate operation \'value\' property is not present, or undefined (applicable in add, replace and test operations)', function () {
                let badFn = function () {
                    let patchOps = [
                        {
                            op: 'add',
                            path: '/age'
                        }
                    ];

                    patch.isValidJsonPatch(patchOps, document);
                };

                expect(badFn).to.throw(Error, 'Operation `value` property is not present (applicable in `add`, `replace` and `test` operations)');
            });

            it('should validate cannot perform an add operation at the desired path', function () {
                let badFn = function () {
                    let patchOps = [
                        {
                            op: 'add',
                            path: '/some/nonexistent/path',
                            value: 'some value'
                        }
                    ];

                    patch.isValidJsonPatch(patchOps, document);
                };

                expect(badFn).to.throw(Error, 'Cannot perform an `add` operation at the desired path');
            });

            it('should validate cannot perform the operation at a path that does not exist', function () {
                let badFn = function () {
                    let patchOps = [
                        {
                            op: 'replace',
                            path: '/some/nonexistent/path',
                            value: 'some value'
                        }
                    ];

                    patch.isValidJsonPatch(patchOps, document);
                };

                expect(badFn).to.throw(Error, 'Cannot perform the operation at a path that does not exist');
            });

            it('should validate cannot perform the operation from a path that does not exist', function () {
                let badFn = function () {
                    let patchOps = [
                        {
                            op: 'move',
                            from: '/some/nonexistent/path',
                            path: '/first_name'
                        }
                    ];

                    patch.isValidJsonPatch(patchOps, document);
                };

                expect(badFn).to.throw(Error, 'Cannot perform the operation from a path that does not exist');
            });

            it('should validate expected an unsigned base-10 integer value, making the new referenced value the array element with the zero-based index', function () {
                let badFn = function () {
                    let patchOps = [
                        {
                            op: 'add',
                            path: '/favorite_colors/a',
                            value: 'violet'
                        }
                    ];

                    patch.isValidJsonPatch(patchOps, document);
                };

                expect(badFn).to.throw(Error, 'Expected an unsigned base-10 integer value, making the new referenced value the array element with the zero-based index');
            });

            it('should validate the specified index MUST NOT be greater than the number of elements in the array', function () {
                let badFn = function () {
                    let patchOps = [
                        {
                            op: 'add',
                            path: '/favorite_colors/3',
                            value: 'yellow'
                        }
                    ];

                    patch.isValidJsonPatch(patchOps, document);
                };

                expect(badFn).to.throw(Error, 'The specified index MUST NOT be greater than the number of elements in the array');
            });

            it('should validate when operation is test and the test fails, applies to applyReducer', function () {
                let badFn = function () {
                    let patchOps = [
                        {
                            op: 'test',
                            path: '/member',
                            value: 'an incorrect value'
                        }
                    ];

                    patch.isValidJsonPatch(patchOps, document);
                };

                expect(badFn).to.throw(Error, 'Test operation failed');
            });
        });

        describe('hasValidValues', function () {

            it('should validate operation \'value\' must not be an empty string', function () {
                let badFn = function () {
                    let patchOps = [
                        {
                            op: 'replace',
                            path: '/first_name',
                            value: ''
                        }
                    ];

                    patch.hasValidValues(patchOps);
                };

                expect(badFn).to.throw(Error, 'Patch value must not be an empty string or whitespace');
            });

            it('should validate operation \'value\' must not only contain whitespace', function () {
                let badFn = function () {
                    let patchOps = [
                        {
                            op: 'replace',
                            path: '/first_name',
                            value: '   '
                        }
                    ];

                    patch.hasValidValues(patchOps);
                };

                expect(badFn).to.throw(Error, 'Patch value must not be an empty string or whitespace');
            });

            it('should validate operation \'value\' must not be a carriage return', function () {
                let badFn = function () {
                    let patchOps = [
                        {
                            op: 'replace',
                            path: '/first_name',
                            value: '\r'
                        }
                    ];

                    patch.hasValidValues(patchOps);
                };

                expect(badFn).to.throw(Error, 'Patch value must not be an empty string or whitespace');
            });

            it('should validate operation \'value\' must not be a newline', function () {
                let badFn = function () {
                    let patchOps = [
                        {
                            op: 'replace',
                            path: '/first_name',
                            value: '\n'
                        }
                    ];

                    patch.hasValidValues(patchOps);
                };

                expect(badFn).to.throw(Error, 'Patch value must not be an empty string or whitespace');
            });
        });

        describe('isNotPatchingCriticalFields', function () {

            it('should validate operation path must not be a partition key', function () {
                let badFn = function () {
                    let patchOps = [
                        {
                            op: 'replace',
                            path: '/last_name',
                            value: 'Geralt'
                        }
                    ];

                    patch.isNotPatchingCriticalFields(patchOps, schema);
                };

                expect(badFn).to.throw(Error, 'Invalid patch target: last_name');
            });
        });
    });

    describe('applyPatchOperations', function() {

        it('should not alter the original document', function() {
            let patchOps = [
                {
                    op: 'replace',
                    path: '/first_name',
                    value: 'Geralt'
                }
            ];

            let cloneDocument = Object.assign({}, document);

            let patchedDocument = patch.applyPatchOperations(patchOps, document);

            expect(patchedDocument.first_name).to.not.equal(cloneDocument.first_name);

            expect(document).to.exist;
            expect(document.person_id).to.equal(cloneDocument.person_id);
            expect(document.first_name).to.equal(cloneDocument.first_name);
            expect(document.last_name).to.equal(cloneDocument.last_name);
            expect(document.favorite_colors).to.equal(cloneDocument.favorite_colors);
        });

        describe('\'add\' operations', function() {

            it('should insert a new value into the array at the specified index', function() {
                let patchOps = [
                    {
                        op: 'add',
                        path: '/favorite_colors/1',
                        value: 'red'
                    }
                ];

                let patchedDocument = patch.applyPatchOperations(patchOps, document);

                expect(patchedDocument).to.exist;
                expect(patchedDocument.favorite_colors).to.exist;
                expect(patchedDocument.favorite_colors.length).to.equal(3);
                expect(patchedDocument.favorite_colors[1]).to.equal('red');
            });

            it('should add a new member to the object if the target location does not already exist', function() {
                let patchOps = [
                    {
                        op: 'add',
                        path: '/age',
                        value: 32
                    }
                ];

                let patchedDocument = patch.applyPatchOperations(patchOps, document);

                expect(patchedDocument).to.exist;
                expect(patchedDocument.age).to.exist;
                expect(patchedDocument.age).to.equal(32);
            });

            it('should replace the value if the target location already exists', function() {
                let patchOps = [
                    {
                        op: 'add',
                        path: '/first_name',
                        value: 'Jane'
                    }
                ];

                let patchedDocument = patch.applyPatchOperations(patchOps, document);

                expect(patchedDocument).to.exist;
                expect(patchedDocument.first_name).to.exist;
                expect(patchedDocument.first_name).to.equal('Jane');
            });
        });

        describe('\'remove\' operations', function() {

            it('should remove the value at the target location', function() {
                let patchOps = [
                    {
                        op: 'remove',
                        path: '/last_name'
                    }
                ];

                let patchedDocument = patch.applyPatchOperations(patchOps, document);

                expect(patchedDocument).to.exist;
                expect(patchedDocument.last_name).to.not.exist;
            });
        });

        describe('\'replace\' operations', function() {

            it('should replace the value at the target location with a new value', function() {
                let patchOps = [
                    {
                        op: 'replace',
                        path: '/first_name',
                        value: 'Jane'
                    }
                ];

                let patchedDocument = patch.applyPatchOperations(patchOps, document);

                expect(patchedDocument).to.exist;
                expect(patchedDocument.first_name).to.exist;
                expect(patchedDocument.first_name).to.equal('Jane');
            });
        });

        describe('\'move\' operations', function() {

            it('should remove the value at the specified location and addit to the target location', function() {
                let patchOps = [
                    {
                        op: 'move',
                        from: '/last_name',
                        path: '/middle_name'
                    }
                ];

                let patchedDocument = patch.applyPatchOperations(patchOps, document);

                expect(patchedDocument).to.exist;
                expect(patchedDocument.middle_name).to.exist;
                expect(patchedDocument.middle_name).to.equal('Doe');
                expect(patchedDocument.last_name).to.not.exist;
            });
        });

        describe('\'copy\' operations', function() {

            it('should copy the value at a specified location to the target location', function() {
                let patchOps = [
                    {
                        op: 'copy',
                        from: '/last_name',
                        path: '/middle_name'
                    }
                ];

                let patchedDocument = patch.applyPatchOperations(patchOps, document);

                expect(patchedDocument).to.exist;
                expect(patchedDocument.middle_name).to.exist;
                expect(patchedDocument.middle_name).to.equal('Doe');
                expect(patchedDocument.last_name).to.exist;
                expect(patchedDocument.last_name).to.equal('Doe');
            });
        });

        describe('\'test\' operations', function() {

            it('should test that a value at the target location is equal to a specified value', function() {
                let patchOps = [
                    {
                        op: 'test',
                        path: '/last_name',
                        value: 'Doe'
                    }
                ];

                let patchedDocument = patch.applyPatchOperations(patchOps, document);

                expect(patchedDocument).to.exist;
                expect(patchedDocument.last_name).to.exist;
                expect(patchedDocument.last_name).to.equal('Doe');
            });

            it('should report test operation failed when the value at the target location is not equal to a specified value', function() {

                let badFn = function() {
                    let patchOps = [
                        {
                            op: 'test',
                            path: '/last_name',
                            value: 'Kerrigan'
                        }
                    ];

                    patch.applyPatchOperations(patchOps, document);
                };
                
                expect(badFn).to.throw(Error, 'Test operation failed');
            });
        });
    });
});